import java.math.BigInteger;

/** Wrapper for {@code java.math.}{@link java.math.BigInteger}.
  * Required for mixed arithmetic with {@code int} or {@code long} operands because {@code java.math.}{@link java.math.BigInteger}
  * does not implement a {@link java.math.BigInteger#valueOf(long) BigInteger(long)} constructor.<br>
  * <br>
  * Note: As class {@code BigIntegerAlgebra} extends class {@code java.math.BigInteger} methods of the later class can be called 
  * on {@code BigIntegerAlgebra} instances. The methods however will not return an {@code BigIntegerAlgebra} instance but an instance of the super 
  * class as result. If needed, call the {@link #BigIntegerAlgebra(BigInteger)} constructor on the result to get a {@code BigIntegerAlgebra}
  * instance again. If this is considered inconvenient, wrapper methods for all relevant methods of {@code BigInteger} need to
  * be implemented in the Algebra wrapper class.
  * @see <a target="_top" href="../programmer_examples/MyFractionAlgebra_UserDefinedAlgebra.html">user defined algebra</a>
  * @see <a target="_top" href="../programmer_examples/ProgrammerExamples.html#biginteger">BigIntegerAlgebra_Demo</a>
  */
public class BigIntegerAlgebra extends BigInteger
{
   static final boolean DEBUG = false;
   
// private BigInteger itsBigInteger = null;
   

   /** Constructs a new BigIntegerAlgebra object from a {@code java.math.BigInteger}. */
   public BigIntegerAlgebra( BigInteger value )
   {
   // itsBigInteger = value;
      super( value.toByteArray() );
   }

   /** Constructs a new BigIntegerAlgebra object. <br>
     * Calls {@link java.math.BigInteger#BigInteger(java.lang.String) java.math.BigInteger(String)} but allows 
     * arbitrary {@code ' '} or {@code '_'} characters in {@code value} to improve readability.
     */
   public BigIntegerAlgebra( String value )
   {
      super( KmgStaticUtilities.squeeze( value, "_ " ) );
   }

   /** Constructs a new BigIntegerAlgebra object. <br>
     * Calls {@link java.math.BigInteger#BigInteger(java.lang.String, int) java.math.BigInteger( value, radix )} 
     * but allows arbitrary {@code ' '} or {@code '_'} characters in {@code value}.
     */
   public BigIntegerAlgebra( String value, int radix )
   {
      super( KmgStaticUtilities.squeeze( value, "_ " ), radix );
   }

   /** Constructs a new BigIntegerAlgebra object. <br>
     * Required for mixed operations: int <i>op</i> BigInteger, BigInteger <i>op</i> int
     */
   public BigIntegerAlgebra( long value )
   {
   // itsBigInteger = BigInteger.valueOf( value );
   // super( "" + value );
      this( BigInteger.valueOf( value ) );
   }


// --- reflection methods ---

// arithmetic methods are required to return BigIntegerAlgebra 

   /** Calls {@link java.math.BigInteger#negate}. */
   public BigIntegerAlgebra neg()
   {
      return new BigIntegerAlgebra( negate() );
   }
   
   /** Adds {@code opnd} to this. <br>
     * Calls {@link java.math.BigInteger#add}.
     */
   public BigIntegerAlgebra add( BigInteger opnd )
   {
      return new BigIntegerAlgebra( super.add( opnd ) );
   }
   
   /** Subtracts {@code opnd} from this. <br>
     * Calls {@link java.math.BigInteger#subtract(java.math.BigInteger)}.
     */
   public BigIntegerAlgebra sub( BigInteger opnd )
   {
      return new BigIntegerAlgebra( subtract( opnd ) );
   }
   
   /** (Post)Multiplies this by {@code opnd}. <br>
     * Calls {@link java.math.BigInteger#multiply}.
     */
   public BigIntegerAlgebra mul( BigInteger opnd )
   {
      return new BigIntegerAlgebra( multiply( opnd ) );
   }

   /** Divides this by {@code opnd}. <br>
     * Calls {@link java.math.BigInteger#divide}.
     */
   public BigIntegerAlgebra div( BigInteger opnd )
   {
      return new BigIntegerAlgebra( divide( opnd ) );
   }
   
   /** Returns the remainder of this divided by {@code opnd}. <br>
     * Calls {@link java.math.BigInteger#remainder}.
     */
   public BigIntegerAlgebra remainder( BigInteger opnd )
   {
      return new BigIntegerAlgebra( super.remainder( opnd ) );
   }
   
   /** Returns the power of {@code exponent} of this. <br>
     * Calls {@link java.math.BigInteger#pow}.
     */
   public BigIntegerAlgebra pow( int exponent )
   {
      return new BigIntegerAlgebra( super.pow( exponent ) );
   }


// --- other reflection methods ---

}
