/** Simple implementation of a fraction algebra (example of a <i>user defined algebra</i>). <br>
  * Stores numerator and denominator with int precision. Allows mixed operations with int and double.<br>
  * For a may be more elaborate implementation 
  * see &nbsp;<a target="_top" href="../programmer_examples/commons_math_examples/0_CommonsMath_README.html">Commons Math</a> &rarr;
  * <a target="_blank" href="http://commons.apache.org/proper/commons-math/javadocs/api-3.6.1/index.html?org/apache/commons/math3/fraction/BigFraction.html"><code>BigFraction</code></a>
  * <!-- <a target="_blank" href="http://commons.apache.org/proper/commons-math/apidocs/index.html?org/apache/commons/math4/fraction/BigFraction.html"><code>BigFraction</code></a>.-->
  * @see <a target="_top" href="../programmer_examples/MyFractionAlgebra_UserDefinedAlgebra.html">user defined algebra</a>
  */
public class MyFractionAlgebra implements JxnCloneableAlgebra, Comparable<MyFractionAlgebra>
{
   static final boolean DEBUG = false;
   
   protected int numerator;
   protected int denominator;
   
   /** Records if the fraction representation with int numerator and denomonator is exact or an approximation.
     * Is internally changed, if during calculations intermedeate results cannot be exactly represented with int numerator and denominator.
     * @see #toString
     */
   protected boolean exact = true;

   /** Constructs a new Fraction object. */
   public MyFractionAlgebra( int numerator, int denominator )
   {
   // if( denominator == 0 ) throw new KmgFormelException( "invalid denominator: " + denominator );
      if( denominator == 0 ) KmgFormelInterpreter.errorExit( "invalid denominator: " + denominator );
      
      this.numerator = numerator;
      this.denominator = denominator;
      reduce();
   }

   /** Equivalent to {@code new MyFractionAlgebra( numerator, 1 )}. <br>
     * Required for mixed operations: int <i>op</i> fraction, fraction <i>op</i> int
     */
   public MyFractionAlgebra( int numerator )
   {
      this( numerator, 1 );
   }

   /** Finds a fraction that represents {@code x}, using the continued fraction algorithm. <br>
     * Required for mixed operations: double <i>op</i> fraction, fraction <i>op</i> double.
     * @see <a target="_blank" href="https://en.wikipedia.org/wiki/Continued_fraction">continued fraction</a>
     */
   public MyFractionAlgebra( double x )
   {
   // boolean DEBUG = true;
      
      if( x == 0 )
      {
         numerator = 0;
         denominator  = 1;
         return;
      }

      // Kettenbruchalgorithmus (continued fraction)
      int nPrev = 0;
      int dPrev = 1;
      double ax = Math.abs(x);
      numerator = 1;
      denominator = 0;
      
      if( DEBUG ) System.out.println( getClass().getName() + "( " + x + " )" );

      while( true )
      {
         if( ax > Integer.MAX_VALUE ) break;
         int ai = (int)Math.floor( ax );
      // if( DEBUG ) System.out.print( " " + ai );
         
         long nNext = ai * (long)numerator;
         long dNext = ai * (long)denominator;
         if( nNext > Integer.MAX_VALUE ) break;
         if( dNext > Integer.MAX_VALUE ) break;

         nNext += nPrev;
         dNext += dPrev;
         if( nNext > Integer.MAX_VALUE ) break;
         if( dNext > Integer.MAX_VALUE ) break;

         nPrev = numerator;
         dPrev = denominator;
         numerator = (int)nNext;
         denominator  = (int)dNext;

         ax %= 1.;
         if( DEBUG ) System.out.println( "  " + ai + ", " + ax );
         if( ax == 0. ) break;
         ax = 1. / ax;
      }
      
      exact = ax == 0.;
      
      if( x < 0. ) numerator = -numerator;
      reduce();  // required?
      if( DEBUG ) System.out.println( "  => " + this );

      if( Math.abs( x - doubleValue() ) > Math.abs(x) * 2./Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "unable to convert " + x + " to " + getClass() + ": " + this );
      }
   }

   /** Returns the internally stored numerator. */
   public int getNumerator()
   {
      return numerator;
   }
   
   /** Returns the internally stored denominator. */
   public int getDenominator()
   {
      return denominator;
   }

   /** Returns, if the fraction (with int numerator and denominator) is exact (otherwise it is an approximation).
     * @see #exact
     */
   public boolean isExact()
   {
      return exact;
   }

   /** Declares that this fraction instance is considered exact.
     * @return this
     * @see #exact
     */
   public MyFractionAlgebra setExact( boolean exact )
   {
      this.exact = exact;
      return this;
   }

   /** Returns this fraction as double. */
   public double doubleValue()
   {
      return (double)numerator / denominator;
   }


// --- interface methods ---

   /** <b>Note:</b> {@code MyFractionAlgebra} implements {@link JxnCloneableAlgebra} only for demonstration.
     * As {@code MyFractionAlgebra} has only 2 int members {@code numerator} and {@code denominator} it
     * should better be implemented as <i>immutable</i> like {@code java.math.}{@link java.math.BigInteger} 
     * or <!-- <a target="_blank" href="http://commons.apache.org/proper/commons-math/userguide/fraction.html">CommonsMath</a>  -->
     * <a target="_top" href="../programmer_examples/commons_math_examples/0_CommonsMath_README.html">Commons Math</a> &rarr;
     * <a target="_blank" href="http://commons.apache.org/proper/commons-math/javadocs/api-3.6.1/index.html?org/apache/commons/math3/fraction/BigFraction.html"><code>BigFraction</code></a>
     * <!-- <a href="http://commons.apache.org/proper/commons-math/apidocs/org/apache/commons/math4/fraction/BigFraction.html"><code>BigFraction</code></a> -->.<br>
     * <br>
     * <b>Description copied from interface: {@link JxnCloneableAlgebra}</b><br>
     * {@inheritDoc}
     */
   public JxnCloneableAlgebra cloneThis()
   {
      MyFractionAlgebra result = new MyFractionAlgebra( numerator, denominator );
      result.exact = exact;
      return result;
   }


// --- reflection methods ---

   /** Adds {@code opnd} to this.
     * @return this
     */
   public MyFractionAlgebra add( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator + denominator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      if( DEBUG ) System.out.println( "add(0): " + this + " " + opnd );

      long g = gcd( denominator, opnd.denominator );
      long n1 = opnd.denominator / g * numerator;
      long n2 = denominator / g * opnd.numerator;
      long num = n1 + n2;
      long den = denominator / g * opnd.denominator;
      if( DEBUG ) System.out.println( "add(1): " + g + " " + n1 + " " + n2 + " -> " + num + " " + den );

      if( Math.abs(n1) > Integer.MAX_VALUE || Math.abs(n2) > Integer.MAX_VALUE 
          || Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
      // throw new KmgFormelException( "integer overflow for: " + this + " (+) " + opnd );
         MyFractionAlgebra tmp = new MyFractionAlgebra( (double)num / den );
         numerator = tmp.numerator;
         denominator = tmp.denominator;
         exact = false;
         return this;
      }

      numerator = (int)num;
      denominator = (int)den;
      reduce();
      exact &= opnd.exact;
      return this;
   }
   
/* Subtracts opnd from this (Implementation not necessary, if constructor MyFractionAlgebra(int) exists).
   public MyFractionAlgebra sub( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator - denominator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      long g = gcd( denominator, opnd.denominator );
      long n1 = opnd.denominator / g * numerator;
      long n2 = denominator / g * opnd.numerator;
      if( Math.abs(n1) > Integer.MAX_VALUE || Math.abs(n2) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (-) " + opnd );
      }
      long num = n1 - n2;
      long den = denominator / g * opnd.denominator;
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (-) " + opnd );
      }
      numerator = (int)num;
      denominator  = (int)den;
      reduce();
      return this;
   }
*/
   
   /** (Post)Multiplies this by {@code opnd}.
     * @return this
     */
   public MyFractionAlgebra mul( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      if( numerator == 0 ) return this;
      if( opnd.numerator == 0 )
      {
         numerator = 0;
         denominator = 1;
         exact &= opnd.exact;
         return this;
      }

      long g1 = gcd( numerator, opnd.denominator );
      long g2 = gcd( denominator, opnd.numerator );
      if( g1 == 0 || g2 == 0 )  // required?
      {
         System.out.println( g1 + " " + g2 );
         numerator = 0;
         denominator = 1;
         return this;
      }

      long num = numerator / g1 * ( opnd.numerator / g2 );
      long den = denominator / g2 * ( opnd.denominator / g1 );
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
      // throw new KmgFormelException( "integer overflow for: " + this + " (*) " + opnd );
         MyFractionAlgebra tmp = new MyFractionAlgebra( (double)num / den );
         numerator = tmp.numerator;
         denominator = tmp.denominator;
         exact = false;
         return this;
      }

      numerator = (int)num;
      denominator = (int)den;
      reduce();  // required?
      exact &= opnd.exact;
      return this;
   }

/* Divides this by opnd (Implementation not necessary, if method inv() exists).
   public MyFractionAlgebra div( MyFractionAlgebra opnd )
   {
      return mul( new MyFractionAlgebra( opnd.denominator, opnd.numerator ) );
   /*
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator;
   // denominator = denominator*((MyFractionAlgebra)opnd).numerator;
      long g1 = gcd( numerator, ((MyFractionAlgebra)opnd).numerator );
      long g2 = gcd( denominator,  ((MyFractionAlgebra)opnd).denominator );
      long num  = numerator / g1 * ( ((MyFractionAlgebra)opnd).denominator  / g2 );
      long den  = denominator  / g2 * ( ((MyFractionAlgebra)opnd).numerator / g1 );
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (/) " + opnd );
      }
      numerator = (int)num;
      denominator = (int)den;
      reduce();
      return this;
   }
   */

   /** Replaces this by its reciprocal.
     * @return this
     */
   public MyFractionAlgebra inv()
   {
      if( numerator == 0 ) KmgFormelInterpreter.errorExit( "invalid numerator: " + numerator );

      int temp = numerator;
      numerator = denominator;
      denominator = temp;
      return this;
   }
   
   /** Replaces this by its power of {@code exponent}.
     * @return this
     */
   public MyFractionAlgebra pow( int exponent )
   {
      double num, den;
      if( exponent > 0 )
      {
         num = Math.pow( numerator, exponent );
         den = Math.pow( denominator, exponent );
      }
      else if( exponent < 0 )
      {
         if( numerator == 0 ) KmgFormelInterpreter.errorExit( "invalid numerator: " + numerator + " for exponent: " + exponent );
         
         num = Math.pow( denominator, -exponent );
         den = Math.pow( numerator, -exponent );
      }
      else  // exponent == 0
      {
         numerator = 1;
         denominator = 1;
         exact = true;
         return this;
      }
      
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         MyFractionAlgebra tmp = new MyFractionAlgebra( (double)num / den );
         numerator   = tmp.numerator;
         denominator = tmp.denominator;
         exact = false;
         return this;
      }

      numerator   = (int)Math.round( num );
      denominator = (int)Math.round( den );
      if( exact ) exact = ( numerator == num ) && ( denominator == den );
   // reduce();  // not necessary
      return this;
   }


// --- overwriten methods ---

/* TODO: equals => gleicher Hashcode
   public int hashCode()
   {
      return ...
   }
*/

   /** Fractions are equal, if both numerator and denominator are equal. <br>
     * (the exactness-information is not required to be equal)
     */
// public boolean equals( MyFractionAlgebra other )  // geht nicht
   public boolean equals( Object other )
   {
      if( numerator   != ((MyFractionAlgebra)other).numerator )   return false;
      if( denominator != ((MyFractionAlgebra)other).denominator ) return false;
      return true;
   }

   /** Returns -1 if &nbsp;{@code this < other}, +1 if &nbsp;{@code other > this}, else 0. */
   public int compareTo( MyFractionAlgebra other )
   {
      double x = doubleValue();
   // double y = ((MyFractionAlgebra)other).doubleValue();
      double y = other.doubleValue();

      if( x < y ) return -1;
      if( x > y ) return +1;
      return 0;
   }

   /** Returns a string representation of the fraction. <br>
     * <br>
     * Shows the fraction as &nbsp;<code>"<i>numerator</i> / <i>denominator</i> = <i>double-value</i>"</code>, if int valued numerator and denominator allow
     * an exact representation of the fraction, e.g. {@code "1 / 8 = 0.125"}. However, if the fraction is an approximation, it is formatted as
     * <code>"<i>double-value</i> ~ <i>numerator</i> / <i>denominator</i>"</code> e.g. {@code "0.3 ~ 3 / 10"} (there is no exact binary representation 
     * of the decimal {@code 0.3}).<br>
     * Note the difference:
     * <pre>
     *    f1 = @{@link #MyFractionAlgebra(int, int) MyFractionAlgebra}( 3, 10 )  ! exact fraction
     *     = 3 / 10 = 0.3 (MyFractionAlgebra)
     *    f2 = @{@link #MyFractionAlgebra(double) MyFractionAlgebra}( 0.3 )  ! approximation of the double {@code 0.3}
     *     = 0.3 ~ 3 / 10 (MyFractionAlgebra)  </pre>   
     * The resulting fraction instances {@code f1} and {@code f2} are equal, but only {@code f1} considers itself exact.
     * A Calculation with fraction instances transfers the exactness-information to its result.
     * @see #exact
     */
   public String toString()
   {
      if( exact ) return numerator + " / " + denominator + " = " + doubleValue();
   // return doubleValue() + " \u2248 " + numerator + " / " + denominator;  // asymp wird zu klein dargestellt
      return doubleValue() + " ~ " + numerator + " / " + denominator;
   }
   
   
// --- other reflection methods ---

   /** Returns the greatest common divisor (gcd) of {@code m} and {@code n}. */
   public static int gcd( int m, int n )
   {
//    if( m > n )
//       return gcd( m - n, n );
//    else if( n > m )
//       return gcd( m, n - m );
//    else
//       return m;

      if( n == 0 ) return 0;

      for(;;)
      {
         int mn = m % n;
         if( mn == 0 ) return Math.abs(n);
         m = n;
         n = mn;
      }
   }
   
   
// --- internal methods ---

   /** Reduces the fraction: devides numerator and denominator by their greatest common divisor. */
   protected void reduce()  // reduce
   {
      int g = gcd( numerator, denominator );
      if ( g == 0 ) return;
      if( denominator < 0 ) g = -g;
      numerator /= g;
      denominator /= g;
   }
   
   /** power.
     * @param exponent must be &gt;= 0
     */
   protected int power( int base, int exponent )  // no more needed
   {
      if( exponent < 0 ) throw new KmgFormelException( "negative exponent not allowed" );
      int exp = exponent;
      
      long result = 1;
      while( exp > 0 )
      {
         result *= base;
         if( Math.abs( result ) > Integer.MAX_VALUE ) throw new KmgFormelException( "integer overflow for: " + this + " ^ " + exponent );
         exp--;
      }
      
      return (int)result;
   }
   
   /** Test. */
   public static void main( String [] args )
   {
      for( int n = 1; n <= Integer.MAX_VALUE; n++ )
      {
         MyFractionAlgebra sqrt_n = new MyFractionAlgebra( Math.sqrt(n) );
         if( sqrt_n.mul( sqrt_n ).equals( new MyFractionAlgebra(n) ) == false )
         {
            System.out.println( n + " " + sqrt_n );  // 1073741825 1.073741824E9 ~ 1073741824 / 1
            break;
         }
         if( n % 10_000_000 == 0 ) System.out.println( n + " " + sqrt_n );
      }
      System.out.println( Integer.MAX_VALUE );
   }
}
