import org.apache.commons.math3.ode.*;

/** Example of a system of first order differential equations (ODEs). <br>
  * <br>
  * Lotka-Volterra equations for rabbits {@code r(t)} and foxes {@code f(t)}:<br>
  * <br>
  * <pre>
  *  dr/dt = r(t) * ( a - b * f(t) )
  *  df/dt = f(t) * ( d * r(t) - c )</pre>
  * @see <a target="_top" href="../0_CommonsMath_README.html#examples">CommonsMath_ODE_Lotka_Volterra~Demo.jxn</a>
  * @see <a target="_blank" href="https://en.wikipedia.org/wiki/Lotka%2dVolterra_equations">&rarr; Wikipedia</a>
  */
public class ODE_Lotka_Volterra implements FirstOrderDifferentialEquations
{
// global parameters:
   double a, b, c, d;

   /** Stores the parameters of the equations. */
   public ODE_Lotka_Volterra( double a, double b, double c, double d )
   {
      this.a = a;
      this.b = b;
      this.c = c;
      this.d = d;
   }
   
// --- interface FirstOrderDifferentialEquations ---

   /** Returns the number of states(=2). */
   public int getDimension()
   {
      return 2;
   }

   /** Computes <code>dr/dt</code> and <code>df/dt</code>. */
   public void computeDerivatives( double t, double [] y, double [] ydot )
   {
   // calculate derivatives
      double rt = y[0];  // rabbits
      double ft = y[1];  // foxes
      ydot[0] = rt * ( a - b * ft );  // dr/dt
      ydot[1] = ft * ( d * rt - c );  // df/dt
   }
}
